/*
 * PureIRCd Internet Relay Chat Daemon, src/s_dconf.c
 * (C) 1999 Carsten Munk (Techie/Stskeeps) <stskeeps@tspre.org>
 *
 * This program is free softwmare; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#define DYNCONF_C
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "config.h"
#include "s_dconf.h"
#include <time.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <utmp.h>
#include <fcntl.h>
#include "h.h"
#if defined( HAVE_STRING_H )
# include <string.h>
#else
# include <strings.h>
# undef strchr
# define strchr index
#endif

#define DoDebug fprintf(stderr, "[%s] %s | %li\n", babuf, __FILE__, __LINE__);
#define AllocCpy(x,y) if ((x) && type == 1) MyFree((x)); x = (char *) MyMalloc(strlen(y) + 1); strcpy(x,y)
#define XtndCpy(x,y) x = (char *) MyMalloc(strlen(y) + 2); *x = '\0'; strcat(x, "*"); strcpy(x,y)

/* internals */
aConfiguration iConf;

aCRline *crlines = NULL;
aBline *blines = NULL;
char *cannotjoin_msg = NULL;

int icx = 0;
char buf[1024];

/* strips \r and \n's from the line */
void iCstrip(char *line)
{
    char *c;

    if ((c = strchr (line, '\n')))
	*c = '\0';
    if ((c = strchr (line, '\r')))
	*c = '\0';
}

/*
 * load_conf()
 */
int load_conf(char *filename, int type)
{
    FILE *zConf;
    char *version = NULL;
    char *i;

    zConf = fopen (filename, "r");

    if (!zConf)
    {
	if (type == 1)
	{
	    sendto_umode(MODE_s, "Could not load %s.", filename);
	    return -1;
	}
	else
	{
	    fprintf(stderr, "* Loading dynamic configuration file (%s)... failed\n", filename);
	    exit(-1);
	}
    }

    i = fgets (buf, 1023, zConf);

    if (!i)
    {
	if (type == 1)
	{
	    sendto_umode(MODE_s, "Error reading from %s.", filename);
	    return -1;
	}
	else
	{
	    fprintf(stderr, "* Loading dynamic configuration file (%s)... failed\n", filename);
	    exit(-1);
	}
    }
    iCstrip (buf);
    version = strtok (buf, "^");
    version = strtok (NULL, "");

    if (!version)
	goto malformed;
    if (!match("2.1.1", version))
    {
	if (strcmp(version, DYNCONF_CONF_VERSION))
	{
	    if (type == 1)
	    {
		sendto_umode(MODE_s, "%s got a non-compatible version (%s)", filename, version);
		return -1;
	    }
	    else
	    {
		fprintf(stderr, "* Error - Incompatible version (%s) in %s - please upgrade\n", version, filename);
		exit (-1);
	    }
	}
	load_conf2 (zConf, filename, type);
	return -1;
    }
    else
    {
malformed:
	if (type == 1)
	{
	    sendto_umode(MODE_s, "Malformed version header in (%s).", filename);
	    return -1;
	}
	else
	{
	    fprintf(stderr, "* Error - Malformed version header in (%s). Please use the correct version ini file.\n", filename);
	    exit(-1);
	}
	return -1;
    }
    return -1;
}

int load_conf2(FILE * conf, char *filename, int type)
{
    char *databuf = buf;
    char *stat = databuf;
    char *p, *setto, *var;

    if (!conf)
	return -1;
    *databuf = '\0';

    while (stat != NULL)
    {
	stat = fgets (buf, 1020, conf);

	if ((*buf == '#') || (*buf == '/'))
	    continue;
	iCstrip (buf);

	if (*buf == '\0')
	    continue;
	p = strtok (buf, " ");

	if (mycmp(p, "set") == 0)
	{
	    var = strtok (NULL, " ");

	    if (var == NULL)
		continue;
	    if (*var == '\0')
		continue;

	    strtok (NULL, " ");
	    setto = strtok (NULL, "");

	    if (strcmp(var, "dyn_shortmotd_state") == 0)
	    {
		DYN_SHORTMOTD_STATE = atoi(setto);
	    }
	    else if (strcmp(var, "dyn_hub_state") == 0)
	    {
		DYN_HUB_STATE = atoi(setto);
	    }
	    else if (strcmp(var, "dyn_rfc1035_anal") == 0)
	    {
		DYN_RFC1035_ANAL = atoi(setto);
	    }
	    else if (strcmp(var, "dyn_no_mixed_case") == 0)
	    {
		DYN_NO_MIXED_CASE = atoi(setto);
	    }


/*	    else if (strcmp(var, "dyn_nojoin_split") == 0)
	    {
		DYN_NOJOIN_SPLIT = atoi(setto);
	    } */
	    else if (strcmp(var, "dyn_allow_links") == 0)
	    {
		DYN_ALLOW_LINKS = atoi(setto);
	    }
	    else if (strcmp(var, "dyn_allow_map") == 0)
	    {
		DYN_ALLOW_MAP = atoi(setto);
	    }
	    else if (strcmp(var, "dyn_identd_comp") == 0)
	    {
		DYN_IDENTD_COMP = atoi(setto);
	    }
	    else if (strcmp(var, "dyn_akg_notices") == 0)
	    {
		DYN_AKG_NOTICES = atoi(setto);
	    }
	    else if (strcmp(var, "dyn_failoper_kill") == 0)
	    {
		DYN_FAILOPER_KILL = atoi(setto);
	    }
	    else if (strcmp(var, "dyn_filter_quitmsg") == 0)
	    {
		DYN_FILTER_QUITMSG = atoi(setto);
	    }
	    else if (strcmp(var, "dyn_proxmon_state") == 0)
	    {
		DYN_PROXMON_STATE = atoi(setto);			
	    }
	    else if (strcmp(var, "dyn_opermotd_state") == 0)
	    {
		DYN_OPERMOTD_STATE = atoi(setto);
	    }
	    else if (strcmp(var, "dyn_network_maxchans") == 0)
	    {
		DYN_NETWORK_MAXCHANS = atol(setto);
	    }
	    else if (strcmp(var, "dyn_hide_ulines") == 0)
	    {
		DYN_HIDE_ULINES = atoi(setto);
	    }
	    else if (strcmp(var, "dyn_nochanops_split") == 0)
	    {
		DYN_NOCHANOPS_SPLIT = atoi(setto);
	    }
	    else if (strcmp(var, "dyn_htmlockon_burst") == 0)
	    {
		DYN_HTMLOCKON_BURST = atoi(setto);
	    }
	    else if (strcmp(var, "dyn_ircii_kludge") == 0)
            {
                DYN_IRCII_KLUDGE = atoi(setto);
            }
	    else if (strcmp(var, "dyn_allow_remove_modex") == 0)
            {
                DYN_ALLOW_REMOVE_MODEX = atoi(setto);
            }
            else if (strcmp(var, "dyn_enable_throttling") == 0)
            {
                DYN_ENABLE_THROTTLING = atoi(setto);
            }
           else if (strcmp(var, "dyn_enable_iline_warn") == 0)
            {
                DYN_ENABLE_ILINE_WARN = atoi(setto);
            }
            else if (strcmp(var, "dyn_enable_malicious_detect") == 0)
            {
                DYN_ENABLE_MALICIOUS_DETECT = atoi(setto);
            }
            else if (strcmp(var, "dyn_whois_wait") == 0)
            {
                WHOIS_WAIT = atoi(setto);
            } 
            else if (strcmp(var, "dyn_connect_modes") == 0)
	    {
		AllocCpy(DYN_CONNECT_MODES, setto);
	    }
	    else if (strcmp(var, "dyn_mailadr_kline") == 0)
	    {
		AllocCpy(DYN_MAILADR_KLINE, setto);
	    }
	    else if (strcmp(var, "dyn_mailadr_gline") == 0)
	    {
		AllocCpy(DYN_MAILADR_GLINE, setto);
	    }
	    else if (strcmp(var, "dyn_proxmon_url") == 0)
	    {
		AllocCpy(DYN_PROXMON_URL, setto);
	    }
	    else if (strcmp(var, "dyn_proxmon_host") == 0)
	    {
		AllocCpy(DYN_PROXMON_HOST, setto);
	    }
	    else if (strcmp(var, "dyn_userchan_join") == 0)
	    {
		AllocCpy(DYN_USERCHAN_JOIN, setto);
	    }
	    else if (strcmp(var, "dyn_operchan_join") == 0)
	    {
		AllocCpy(DYN_OPERCHAN_JOIN, setto);
	    }
	    else if (strcmp(var, "dyn_operhost_netadmin") == 0)
	    {
		AllocCpy(DYN_OPERHOST_NETADMIN, setto);
	    }
	    else if (strcmp(var, "dyn_operhost_techadmin") == 0)
	    {
		AllocCpy(DYN_OPERHOST_TECHADMIN, setto);
	    }
	    else if (strcmp(var, "dyn_operhost_sadmin") == 0)
	    {
		AllocCpy(DYN_OPERHOST_SADMIN, setto);
	    }
	    else if (strcmp(var, "dyn_operhost_abuseadmin") == 0)
	    {
		AllocCpy(DYN_OPERHOST_ABUSEADMIN, setto);
	    }
	    else if (strcmp(var, "dyn_operhost_admin") == 0)
	    {
		AllocCpy(DYN_OPERHOST_ADMIN, setto);
	    }
	    else if (strcmp(var, "dyn_operhost_coadmin") == 0)
	    {
		AllocCpy(DYN_OPERHOST_COADMIN, setto);
	    }
	    else if (strcmp(var, "dyn_operhost_globop") == 0)
	    {
		AllocCpy(DYN_OPERHOST_GLOBOP, setto);
	    }
	    else if (strcmp(var, "dyn_operhost_locop") == 0)
	    {
		AllocCpy(DYN_OPERHOST_LOCOP, setto);
	    }
	    else if (strcmp(var, "dyn_network_netname") == 0)
	    {
		AllocCpy(DYN_NETWORK_NETNAME, setto);
	    }
	    else if (strcmp(var, "dyn_server_services") == 0)
	    {
		AllocCpy(DYN_SERVER_SERVICES, setto);
	    }
	    else if (strcmp(var, "dyn_malicious_1") == 0)
            {
                AllocCpy(DYN_MALICIOUS_1, setto);
            }
	    else if (strcmp(var, "dyn_malicious_2") == 0)
            {
                AllocCpy(DYN_MALICIOUS_2, setto);
            }
	    else if (strcmp(var, "dyn_malicious_3") == 0)
            {
                AllocCpy(DYN_MALICIOUS_3, setto);
            }
	    else
	    {
	    if (type == 0)
		fprintf(stderr, "Dynconf error: Invalid Set [%s %s] setting in %s \n", var, setto, filename);
	    else
		sendto_umode(MODE_s, "conf/ircd.ini warning: Invalid set [%s %s] setting in %s", var, setto, filename);
	    }
	}
	else
	{
	if (type == 0)
	{
	    fprintf(stderr, "* Error: [%s] in (%s) is not a known setting\n", p, filename);
	}
	else
	{
	    sendto_umode(MODE_s, "conf/ircd.ini warning: [%s] in (%s) is not a known setting", p, filename);
	}
     }
  }
  if (type == 0)
  {
      fprintf(stderr, "* Loading dynamic configuration file (%s)... done\n", filename);
  }
  else
  {
      sendto_umode(MODE_s, "Loaded %s ...", filename);
  }
  return 0;
}

void init_dynconf (void)
{
    bzero (&iConf, sizeof (iConf));
}

void check_dynconf(int type)
{
    char *errormsg = MyMalloc(16384);
    *errormsg = '\0';  

    if (!DYN_MAILADR_KLINE || (*DYN_MAILADR_KLINE == '\0'))
    {
	if (type == 0)
	    strcat(errormsg, "- Missing dyn_mailadr_kline\n");		
	else
	    sendto_umode(MODE_s, "Dynconf ERROR: Set dyn_mailadr_kline in ircd.ini is invalid (NULL value).");
    }
    if (!DYN_CONNECT_MODES || (*DYN_CONNECT_MODES == '\0'))
	strcat(errormsg, "* Missing DYN_CONNECT_MODES\n");
    if (DYN_CONNECT_MODES)
    {
	if (!strchr(DYN_CONNECT_MODES, 'x'))
	{
	    strcat(errormsg, "* Missing x in dyn_connect_modes\n");
	}
    }
    if (*errormsg != '\0')
    {
	fprintf(stderr, "* ERRORS IN DYNAMIC CONFIGURATION (listed below):\n");
	fprintf(stderr, errormsg);
	MyFree(errormsg);
	exit(-1);
    }
    MyFree(errormsg);
}

/*
 * channel_canjoin()
 */
int channel_canjoin(aClient *sptr, char *name)
{
    aCRline *p;

    if (IsAnOper(sptr))
	return 1;
    if (IsULine(sptr))
	return 1;
    if (!crlines)
	return 1;
    for (p = crlines; p; p = p->next)
    {
	if (!match (p->channel, name))
	    return 1;
    }
    return 0;
}

/*
 * cr_add()
 */
int cr_add(char *channel, int type)
{
    aCRline *fl;
    fl = (aCRline *) MyMalloc (sizeof (aCRline));

    AllocCpy(fl->channel, channel);
    fl->type = type;
    fl->next = crlines;
    fl->prev = NULL;

    if (crlines)
	crlines->prev = fl;
    crlines = fl;
    return 0;
}

aCRline *cr_del(fl)
    aCRline *fl;
{
    aCRline *p, *q;

    for (p = crlines; p; p = p->next)
    {
	if (p == fl)
	{
	    q = p->next;
	    MyFree ((char *) p->channel);

	    if (p->prev)
		p->prev->next = p->next;
	    else
		crlines = p->next;
	    if (p->next)
		p->next->prev = p->prev;
	    MyFree ((aCRline *) p);
	    return q;
	}
    }
    return NULL;
}

/* 
 * cr_loadconf()
 */
int cr_loadconf(char *filename, int type)
{
    char buf[2048];
    char *x, *y;
    FILE *f;

    f = fopen (filename, "r");

    if (!f)
	return -1;

    while (fgets(buf, 2048, f))
    {
	if (buf[0] == '#' || buf[0] == '/' || buf[0] == '\0' || buf[0] == '\n' || buf[0] == '\r')
	    continue;
	iCstrip (buf);
	x = strtok (buf, " ");

	if (mycmp(x, "allow") == 0)
	{
	    y = strtok(NULL, " ");
	    if (!y)
		continue;
	    cr_add (y, 0);
	}
	else if (mycmp(x, "msg") == 0)
	{
	    y = strtok(NULL, "");

	    if (!y)
		continue;
	    if (cannotjoin_msg)
		MyFree ((char *) cannotjoin_msg);
	    cannotjoin_msg = MyMalloc (strlen (y) + 1);
	    strcpy(cannotjoin_msg, y);
	}
    }
    return 0;
}

/*
 * cr_rehash()
 */
void cr_rehash(void)
{
    aCRline *p, q;

    for (p = crlines; p; p = p->next)
    {
	if ((p->type == 0) || (p->type == 2))
	{
	    q.next = cr_del (p);
	    p = &q;
	}
    }
    cr_loadconf(IRCD_RESTRICT, 1);
}

/*
 * cr_report()
 */
void cr_report(aClient *sptr)
{
    aCRline *tmp;
    char *filemask;

    if (!crlines)
    {
	sendto_one(sptr, ":%s %d %s :There are currently no channel restrictions in place on %s", me.name, RPL_SETTINGS, sptr->name, me.name);
	return;
    }
    sendto_one(sptr, ":%s %d %s :*** Allowed Channels ***", me.name, RPL_SETTINGS, sptr->name);
    sendto_one(sptr, ":%s %d %s :", me.name, RPL_SETTINGS, sptr->name);

    for (tmp = crlines; tmp; tmp = tmp->next)
    {
	filemask = BadPtr (tmp->channel) ? "<NULL>" : tmp->channel;
	sendto_one(sptr, ":%s %d %s :%s", me.name, RPL_SETTINGS, sptr->name, filemask);

    }
    sendto_one(sptr, ":%s %d %s :", me.name, RPL_SETTINGS, sptr->name);
}

/*
 * dccdeny stuff
 */
aBline *dcc_isforbidden(aClient *cptr, aClient *sptr, aClient *target, char *filename)
{
    aBline *p;

    if (!blines || !target || !filename)
	return NULL;
    if (HasUmodeo(sptr) || IsULine(sptr))
	return NULL;
    if (HasUmodeo(target))
	return NULL;

    for (p = blines; p; p = p->next)
    {
	if (!match(p->mask, filename))
	    return p;
    }
    return NULL;
}

int dcc_add_bline(char *mask, char *reason, int type)
{
    aBline *bl;
    bl = (aBline *) MyMalloc(sizeof(aBline));

    AllocCpy(bl->mask, mask);
    AllocCpy(bl->reason, reason);
    bl->type = type;
    bl->next = blines;
    bl->prev = NULL;

    if (blines)
	blines->prev = bl;
    blines = bl;
    return 0;
}

aBline *dcc_del_bline(bl)
    aBline *bl;
{
    aBline *p, *q;

    for (p = blines; p; p = p->next)
    {
	if (p == bl)
	{
	    q = p->next;
	    MyFree((char *)p->mask);
	    MyFree((char *)p->reason);

	    if (p->prev)
		p->prev->next = p->next;
	    else
		blines = p->next;
	    if (p->next)
		p->next->prev = p->prev;
	    MyFree((aBline *) p);
	    return q;
	}
    }
    return NULL;
}

void dcc_wipe_all(void)
{
    aBline *p, q;

    for (p = blines; p; p = p->next)
    {
	q.next = dcc_del_bline(p);
	p = &q;
    }
}

int dcc_del_wild_match(char *mask)
{
    aBline *p;
    int found = 0;

    for (p = blines; p; p = p->next)
    {
	if (!match(mask, p->mask) && p->type == 2)
	{
	    found = 1;
	    dcc_del_bline(p);
	}
    }
    return found;
}

aBline *dcc_find(char *mask)
{
    aBline *p;

    for (p = blines; p; p = p->next)
    {
	if (!strcmp(p->mask, mask))
	    return (p);
    }
    return NULL;
}

void send_blines(aClient *sptr)
{
    aBline *p;

    for (p = blines; p; p = p->next)
    {
	if (p->type == 1)
	    sendto_one(sptr, ":%s SVSBLINE + %s :%s", me.name, p->mask, p->reason);
    }
}

void dcc_rehash(void)
{
    aBline *p, q;

    for (p = blines; p; p = p->next)
    {
	if ((p->type == 0) || (p->type == 2))
	{
	    q.next = dcc_del_bline(p);
	    p = &q;
	}
    }
    dcc_loadconf(IRCD_DCCDENY, 1);
}

void dcc_wipe_services(void)
{
    aBline *p, q;

    for (p = blines; p; p = p->next)
    {
	if ((p->type == 1))
	{
	    q.next = dcc_del_bline(p);
	    p = &q;
	}
    }
}

void report_blines(aClient *sptr)
{
    aBline *tmp;
    char *filemask, *reason;
    char a;
    a = 0;

    if (!blines)
    {
	sendto_one(sptr, ":%s %d %s :There are currently no B:Lines in place on %s", me.name, RPL_SETTINGS, sptr->name, me.name);
	return;
    }
    sendto_one(sptr, ":%s %d %s :*** Blocked Files/Filetypes ***", me.name, RPL_SETTINGS, sptr->name);

    for (tmp = blines; tmp; tmp = tmp->next)
    {
	filemask = BadPtr(tmp->mask) ? "<NULL>" : tmp->mask;
	reason = BadPtr(tmp->reason) ? "<NULL>" : tmp->reason;
	if (tmp->type == 0)
	    a = 'c';
	if (tmp->type == 1)
	    a = 's';
	if (tmp->type == 2)
	    a = 'o';
	sendto_one(sptr, ":%s %i %s :(%c) %s %s", me.name, RPL_TEXT, sptr->name, a, filemask, reason);
    }
}

/* 
 * dcc_loadconf()
 */
int dcc_loadconf(char *filename, int type)
{
    char buf[2048];
    char *x, *y, *z;
    FILE *f;

    f = fopen (filename, "r");

    if (!f)
	return -1;
    while (fgets(buf, 2048, f))
    {
	if (buf[0] == '#' || buf[0] == '/' || buf[0] == '\0' || buf[0] == '\n' || buf[0] == '\r')
	    continue;
	iCstrip(buf);
	x = strtok(buf, " ");

	if (mycmp(x, "deny") == 0)
	{
	    y = strtok(NULL, " ");
	    z = strtok(NULL, "");

	    if (!z)
		continue;
	    if (!dcc_find(y))
		dcc_add_bline(y, z, 0);
	}
    }
    return 0;
}

/* 
 * m_dccdeny()
 *  parv[0] = sender
 *  parv[1] = file
 *  parv[2] = reason
 */
int m_dccdeny(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    if (!MyClient(sptr))
	return 0;
    if (!HasUmodeA(sptr))
    {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }
    if (parc < 2)
    {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "DCCDENY");
	return 0;
    }
    if (BadPtr(parv[2]))
    {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "DCCDENY");
	return 0;
    }
    if (!dcc_find(parv[1]))
    {
	sendto_umode(MODE_s, "%s added a temp dccdeny for %s (%s)", parv[0], parv[1], parv[2]);
	dcc_add_bline(parv[1], parv[2], 2);
    }
    else
	sendto_one(sptr, "NOTICE %s :*** Notice -- %s already has a dccdeny", parv[0], parv[1]);
    return 0;
}

/*
 * m_undccdeny()
 *  parv[0] = sender
 *  parv[1] = file/mask
 */
int m_undccdeny(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aBline *p;

    if (!MyClient(sptr))
	return 0;
    if (!HasUmodeA(sptr))
    {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }
    if (parc < 2)
    {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "UNDCCDENY");
	return 0;
    }
    if (BadPtr(parv[1]))
    {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "UNDCCDENY");
	return 0;
    }
    if ((p = dcc_find(parv[1])) && p->type == 2)
    {
	sendto_umode(MODE_s, "%s removed a temp dccdeny for %s", parv[0], parv[1]);
	dcc_del_bline(p);
    }
    else if (dcc_del_wild_match(parv[1]) == 1)
	sendto_umode(MODE_s, "%s removed a temp dccdeny for all dccdenys matching %s", parv[0], parv[1]);
    else
	sendto_one(sptr, "NOTICE %s :*** Notice -- Unable to find a temp dccdeny matching %s", parv[0], parv[1]);
    return 0;
}

/*
 * m_svsbline()
 */
int m_svsbline(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    if (!IsServer(sptr))
	return 0;
    if (parc < 2)
	return 0;
    if (!IsULine(sptr)) 
    {
	sendto_serv_butone(&me,":%s GLOBOPS :Non U:Lined server %s trying to SVSBLINE!", me.name, sptr->name);
	send_globops("from %s: Non U:Lined server %s trying to SVSBLINE!", me.name, sptr->name);
	return 0;
    }

    switch (*parv[1])
    {
	case '+':
	{
	    if (parc < 4)
		return 0;
	    if (!dcc_find(parv[2]))
		dcc_add_bline(parv[2], parv[3], 1);
	    if (IsULine(sptr))
		sendto_serv_butone(cptr, ":%s SVSBLINE + %s :%s", sptr->name, parv[2], parv[3]);
	    break;
	}
	case '-':
	{
	    if (!IsULine(sptr))
		return 0;
	    if (parc < 3)
		return 0;
	    dcc_del_bline(dcc_find(parv[2]));
	    sendto_serv_butone(cptr, ":%s SVSBLINE - %s", sptr->name, parv[2]);
	    break;
	}
	case '*':
	{
	    if (!IsULine(sptr))
		return 0;
	    dcc_wipe_services();
	    sendto_serv_butone(cptr, ":%s SVSBLINE *", sptr->name);
	    break;
	}
    }
    return 0;
}
